@@ -24,7 +24,9 @@ import binascii
 import logging
 import time
 from array import array
+from contextlib import contextmanager
 from struct import *
+from typing import Optional
 
 import usb
 
@@ -57,7 +59,14 @@ def icc_compose(msg_type, data_len, slot, seq, param, data):
     return pack("<BiBBBH", msg_type, data_len, slot, seq, 0, param) + data
 
 
-def iso7816_compose(ins, p1, p2, data, cls=0x00, le=None):
+def iso7816_compose(
+    ins: int,
+    p1: int,
+    p2: int,
+    data: bytes = b"",
+    cls: int = 0x00,
+    le: Optional[int] = None,
+) -> bytes:
     data_len = len(data)
     if data_len == 0:
         if not le:
@@ -65,10 +74,22 @@ def iso7816_compose(ins, p1, p2, data, cls=0x00, le=None):
         else:
             return pack(">BBBBB", cls, ins, p1, p2, le)
     else:
-        if not le:
-            return pack(">BBBBB", cls, ins, p1, p2, data_len) + data
+        if data_len <= 255:
+            if not le:
+                return pack(">BBBBB", cls, ins, p1, p2, data_len) + data
+            else:
+                return (
+                    pack(">BBBBB", cls, ins, p1, p2, data_len) + data + pack(">B", le)
+                )
         else:
-            return pack(">BBBBB", cls, ins, p1, p2, data_len) + data + pack(">B", le)
+            if not le:
+                return pack(">BBBBBH", cls, ins, p1, p2, 0, data_len) + data
+            else:
+                return (
+                    pack(">BBBBBH", cls, ins, p1, p2, 0, data_len)
+                    + data
+                    + pack(">B", le)
+                )
 
 
 # This class only supports Gnuk (for now)
@@ -133,6 +154,17 @@ class gnuk_token(object):
     def release_gnuk(self):
         self.__devhandle.releaseInterface()
 
+    @contextmanager
+    def release_on_exit(self, *args, **kwds):
+        """
+        Release GNUK device once finished, to not block other smart card services
+        trying to claim it
+        """
+        try:
+            yield self
+        finally:
+            self.release_gnuk()
+
     def stop_gnuk(self):
         self.__devhandle.releaseInterface()
         if self.__hid_intf:
@@ -705,7 +737,10 @@ def gnuk_devices_by_vidpid():
                 break
 
 
-def get_gnuk_device(verbose=True, logger: logging.Logger = None):
+def get_gnuk_device(verbose=True, logger: Optional[logging.Logger] = None):
+    from usb import USBError
+
+    candidates = []
     icc = None
     for (dev, config, intf) in gnuk_devices():
         try:
@@ -725,8 +760,13 @@ def get_gnuk_device(verbose=True, logger: logging.Logger = None):
                         f'Device: name: "{dev.filename}", c/i: {config.value}/{intf.interfaceNumber}'
                     )
             break
-        except:
+        except USBError as e:
+            # USBError(16, 'Resource busy')
+            candidates.append(icc)
+        except Exception:
             pass
+    if not icc and candidates:
+        raise OnlyBusyICCError(f"Found only busy ICC: {candidates}")
     if not icc:
         raise ValueError("No ICC present")
     status = icc.icc_get_status()
@@ -794,3 +834,7 @@ def parse_kdf_data(kdf_data):
             raise ValueError("data does not much")
         hash_admin = kdf_data[78:110]
     return (algo, subalgo, iters, salt, salt_reset, salt_admin, hash_user, hash_admin)
+
+
+class OnlyBusyICCError(ValueError):
+    pass
